unit uCapturableWebBrowser;

interface

uses
  Windows, Messages, Classes, Graphics, StdCtrls, ShDocVw, ActiveX, OleCtrls;

type
  TCapturableWebBrowserOption = (
    woScript,             // XNvg̎s
    woImage,              // C[W̕\
    woTheme,              // e[}
    wo3D,                 // 3D \
    woAllowDrop);         // hbv

  TCapturableWebBrowserOptions = set of TCapturableWebBrowserOption;

  PDocHostUIInfo = ^TDocHostUIInfo;
  TDocHostUIInfo = packed record
    cbSize: ULONG;
    dwFlags: DWORD;
    dwDoubleClick: DWORD;
  end;

  IDocHostUIHandler = interface(IUnknown)
    ['{bd3f23c0-d43e-11cf-893b-00aa00bdce1a}']
    function ShowContextMenu(
      const constdwID: DWORD;
      const ppt: PPoint;
      const pcmdtReserved: IUnknown;
      const pdispReserved: IDispatch): HRESULT; stdcall;
    function GetHostInfo(var pInfo: TDocHostUIInfo): HRESULT; stdcall;
    function ShowUI(
      const constdwID: DWORD;
      const pActiveObject: IOleInPlaceActiveObject;
      const pCommandTarget: IOleCommandTarget;
      const pFrame: IOleInPlaceFrame;
      const pDoc: IOleInPlaceUIWindow): HRESULT; stdcall;
    function HideUI: HRESULT; stdcall;
    function UpdateUI: HRESULT; stdcall;
    function EnableModeless(const fEnable: BOOL): HRESULT; stdcall;
    function OnDocWindowActivate(const fActivate: BOOL): HRESULT; stdcall;
    function OnFrameWindowActivate(const fActivate: BOOL): HRESULT; stdcall;
    function ResizeBorder(
      const prcBorder: PRect;
      const pUIWindow: IOleInPlaceUIWindow;
      const fRameWindow: BOOL): HRESULT; stdcall;
    function DocHostTranslateAccelerator(
      const lpMsg: PMsg;
      const pguidCmdGroup: PGUID;
      const nCmdID: DWORD): HRESULT; stdcall;
    function GetOptionKeyPath(
      out pchKey: POleStr;
      const dwReserved: DWORD): HRESULT; stdcall;
    function GetDropTarget(
      const pDropTarget: IDropTarget;
      out ppDropTarget: IDropTarget): HRESULT; stdcall;
    function GetExternal(out ppDispatch: IDispatch): HRESULT; stdcall;
    function TranslateUrl(
      const dwTranslate: DWORD;
      const pchURLIn: POleStr;
      out ppchURLOut: POleStr): HRESULT; stdcall;
    function FilterDataObject(
      const pDO: IDataObject;
      out ppDORet: IDataObject): HRESULT; stdcall;
  end;

  TCapturableWebBrowser = class(TWebBrowser, IDocHostUIHandler)
  private
    FOptions: TCapturableWebBrowserOptions;
    FScrollBars: TScrollStyle;
    procedure SetScrollBars(const Value: TScrollStyle);
  protected
    function IstBody: Boolean;
    // IDocHostUIHandler
    function ShowContextMenu(
      const dwID: DWORD;
      const ppt: PPoint;
      const pcmdtReserved: IUnknown;
      const pdispReserved: IDispatch): HRESULT; stdcall;
    function GetHostInfo(var pInfo: TDocHostUIInfo): HRESULT; stdcall;
    function ShowUI(
      const dwID: DWORD;
      const pActiveObject: IOleInPlaceActiveObject;
      const pCommandTarget: IOleCommandTarget;
      const pFrame: IOleInPlaceFrame;
      const pDoc: IOleInPlaceUIWindow): HRESULT; stdcall;
    function HideUI: HRESULT; stdcall;
    function UpdateUI: HRESULT; stdcall;
    function EnableModeless(const fEnable: BOOL): HRESULT; stdcall;
    function OnDocWindowActivate(const fActivate: BOOL): HRESULT; stdcall;
    function OnFrameWindowActivate(const fActivate: BOOL): HRESULT; stdcall;
    function ResizeBorder(
      const prcBorder: PRect;
      const pUIWindow: IOleInPlaceUIWindow;
      const fRameWindow: BOOL): HRESULT; stdcall;
    function DocHostTranslateAccelerator(
      const lpMsg: PMsg;
      const pguidCmdGroup: PGUID;
      const nCmdID: DWORD): HRESULT; stdcall;
    function GetOptionKeyPath(
      out pchKey: POleStr;
      const dwReserved: DWORD): HRESULT; stdcall;
    function GetDropTarget(
      const pDropTarget: IDropTarget;
      out ppDropTarget: IDropTarget): HRESULT; stdcall;
    function GetExternal(out ppDispatch: IDispatch): HRESULT; stdcall;
    function TranslateUrl(
      const dwTranslate: DWORD;
      const pchURLIn: POleStr;
      out ppchURLOut: POleStr): HRESULT; stdcall;
    function FilterDataObject(
      const pDO: IDataObject;
      out ppDORet: IDataObject): HRESULT; stdcall;
  public
    constructor Create(vOwner: TComponent); override;
    destructor Destroy; override;
    procedure LoadFromFile(const vFileName: String);
    procedure LoadFromStream(const vStream: TStream);
    procedure PaintTo(vBmp: TBitmap; vWidth: Integer = 0; vHeight: Integer = 0);
    procedure Operate(vCommand: String);
    procedure Copy;
    procedure SelectAll;
  published
    property Options: TCapturableWebBrowserOptions read FOptions write FOptions;
    property ScrollBars: TScrollStyle read FScrollBars write SetScrollBars;
  end;

procedure Register;

implementation

uses
  Controls, SysUtils, Forms, Menus, MSHTML;

const
  DOCHOSTUIFLAG_DIALOG = $00000001;
  DOCHOSTUIFLAG_DISABLE_HELP_MENU = $00000002;
  DOCHOSTUIFLAG_NO3DBORDER = $00000004;
  DOCHOSTUIFLAG_SCROLL_NO = $00000008;
  DOCHOSTUIFLAG_DISABLE_SCRIPT_INACTIVE = $00000010;
  DOCHOSTUIFLAG_OPENNEWWIN = $00000020;
  DOCHOSTUIFLAG_DISABLE_OFFSCREEN = $00000040;
  DOCHOSTUIFLAG_FLAT_SCROLLBAR = $00000080;
  DOCHOSTUIFLAG_DIV_BLOCKDEFAULT = $00000100;
  DOCHOSTUIFLAG_ACTIVATE_CLIENTHIT_ONLY = $00000200;
  DOCHOSTUIFLAG_OVERRIDEBEHAVIORFACTORY = $00000400;
  DOCHOSTUIFLAG_CODEPAGELINKEDFONTS = $00000800;
  DOCHOSTUIFLAG_URL_ENCODING_DISABLE_UTF8 = $00001000;
  DOCHOSTUIFLAG_URL_ENCODING_ENABLE_UTF8 = $00002000;
  DOCHOSTUIFLAG_ENABLE_FORMS_AUTOCOMPLETE = $00004000;
  DOCHOSTUIFLAG_ENABLE_INPLACE_NAVIGATION = $00010000;
  DOCHOSTUIFLAG_IME_ENABLE_RECONVERSION = $00020000;
  DOCHOSTUIFLAG_THEME = $00040000;
  DOCHOSTUIFLAG_NOTHEME = $00080000;
  DOCHOSTUIFLAG_NOPICS = $00100000;

  DOCHOSTUIDBLCLK_DEFAULT = 0;
  DOCHOSTUIDBLCLK_SHOWPROPERTIES = 1;
  DOCHOSTUIDBLCLK_SHOWCODE = 2;

type
  ICustomDoc = interface(IUnknown)
    ['{3050f3f0-98b5-11cf-bb82-00aa00bdce0b}']
    function SetUIHandler
      (const pUIHandler: IDocHostUIHandler): HRESULT; stdcall;
  end;

procedure Register;
begin
  RegisterComponents('FuRinKaZan', [TCapturableWebBrowser]);
end;

{ TCapturableWebBrowser }

constructor TCapturableWebBrowser.Create(vOwner: TComponent);
begin
  inherited;

  FOptions := [woImage, wo3D];
end;

destructor TCapturableWebBrowser.Destroy;
var
  Doc: IDispatch;
  CustDoc: ICustomDoc;
begin
  Doc := Document;

  if (Doc <> nil) and (Doc.QueryInterface(ICustomDoc, CustDoc) = S_OK) then
    CustDoc.SetUIHandler(nil);

  inherited;
end;

function TCapturableWebBrowser.FilterDataObject(
  const pDO: IDataObject;
  out ppDORet: IDataObject): HRESULT;
begin
  Result := E_NOTIMPL;
end;

function TCapturableWebBrowser.GetDropTarget(
  const pDropTarget: IDropTarget;
  out ppDropTarget: IDropTarget): HRESULT;
begin
  if (woAllowDrop in FOptions) then
    Result := S_OK
  else
    Result := S_FALSE;
end;

function TCapturableWebBrowser.GetExternal(out ppDispatch: IDispatch): HRESULT;
begin
  Result := E_NOTIMPL;
end;

function TCapturableWebBrowser.GetHostInfo(var pInfo: TDocHostUIInfo): HRESULT;
var
  Flag: DWORD;
begin
  Flag := 0;

  if not (woScript in FOptions) then
    Flag := Flag or DOCHOSTUIFLAG_DISABLE_SCRIPT_INACTIVE;

  if not (woImage in FOptions) then
    Flag := Flag or DOCHOSTUIFLAG_NOPICS;

  if (woTheme in FOptions) then
    Flag := Flag or DOCHOSTUIFLAG_THEME;

  if not (wo3D in FOptions) then
    Flag := Flag or DOCHOSTUIFLAG_NO3DBORDER or DOCHOSTUIFLAG_FLAT_SCROLLBAR;

  if (FScrollBars = ssNone) then
    Flag := Flag or DOCHOSTUIFLAG_SCROLL_NO;

  with pInfo do begin
    cbSize := SizeOf(pInfo);
    dwFlags := Flag;
    dwDoubleClick := DOCHOSTUIDBLCLK_DEFAULT;
  end;

  Result := S_OK;
end;

function TCapturableWebBrowser.GetOptionKeyPath(
  out pchKey: POleStr;
  const dwReserved: DWORD): HRESULT;
begin
  Result := E_NOTIMPL;
end;

function TCapturableWebBrowser.HideUI: HRESULT;
begin
  Result := E_NOTIMPL;
end;

function TCapturableWebBrowser.IstBody: Boolean;
begin
  Result :=
    (IUnknown(OleObject.document) <> nil) and
    (IUnknown(OleObject.document.body) <> nil)
end;

procedure TCapturableWebBrowser.LoadFromFile(const vFileName: String);
begin
  if (not FileExists(vFileName)) then
    Exit;

  Navigate(vFileName);
end;

procedure TCapturableWebBrowser.LoadFromStream(const vStream: TStream);
var
  StreamInit: IPersistStreamInit;
begin
  if
    (Document <> nil) and
    (Document.QueryInterface(IPersistStreamInit, StreamInit) = S_OK)
  then begin
    StreamInit.InitNew;
    StreamInit.Load(TStreamAdapter.Create(vStream));
  end;
end;

function TCapturableWebBrowser.OnDocWindowActivate(
  const fActivate: BOOL): HRESULT;
begin
  Result := E_NOTIMPL;
end;

function TCapturableWebBrowser.OnFrameWindowActivate(
  const fActivate: BOOL): HRESULT;
begin
  Result := E_NOTIMPL;
end;

function TCapturableWebBrowser.ResizeBorder(
  const prcBorder: PRect;
  const pUIWindow: IOleInPlaceUIWindow;
  const fRameWindow: BOOL): HRESULT;
begin
  Result := E_NOTIMPL;
end;

function TCapturableWebBrowser.ShowContextMenu(
  const dwID: DWORD;
  const ppt: PPoint;
  const pcmdtReserved: IUnknown;
  const pdispReserved: IDispatch): HRESULT;
begin
  Result := E_NOTIMPL;
end;

function TCapturableWebBrowser.ShowUI(
  const dwID: DWORD;
  const pActiveObject: IOleInPlaceActiveObject;
  const pCommandTarget: IOleCommandTarget;
  const pFrame: IOleInPlaceFrame;
  const pDoc: IOleInPlaceUIWindow): HRESULT;
begin
  if (pFrame <> nil) then
    pFrame.SetActiveObject(pActiveObject, nil);

  Result := S_FALSE;
end;

function TCapturableWebBrowser.DocHostTranslateAccelerator(
  const lpMsg: PMsg;
  const pguidCmdGroup: PGUID;
  const nCmdID: DWORD): HRESULT;
begin
  Result := E_NOTIMPL;
end;

function TCapturableWebBrowser.EnableModeless(const fEnable: BOOL): HRESULT;
begin
  Result := E_NOTIMPL;
end;

function TCapturableWebBrowser.TranslateUrl(
  const dwTranslate: DWORD;
  const pchURLIn: POleStr;
  out ppchURLOut: POleStr): HRESULT;
begin
  Result := E_NOTIMPL;
end;

function TCapturableWebBrowser.UpdateUI: HRESULT;
begin
  Result := E_NOTIMPL;
end;

procedure TCapturableWebBrowser.SetScrollBars(const Value: TScrollStyle);
begin
  if (FScrollBars = Value) then
    Exit;

  FScrollBars := Value;
  RecreateWnd;
end;

procedure TCapturableWebBrowser.Operate(vCommand: String);
begin
  if (IstBody) then
    try
      OleObject.document.body.createTextRange.execCommand(vCommand);
    except
    end;
end;

procedure TCapturableWebBrowser.Copy;
begin
  Operate('Copy');
end;

procedure TCapturableWebBrowser.SelectAll;
begin
  Operate('SelectAll');
end;

procedure TCapturableWebBrowser.PaintTo(
  vBmp: TBitmap;
  vWidth: Integer = 0;
  vHeight: Integer = 0);
var
  KeepAlign: TAlign;
  OrgWidth, OrgHeight: Integer;
  X, Y: Integer;
  P, Q: Integer;
  S, T: Integer;
  OrgPos: TPoint;
  HTMLDoc: IHTMLDocument2;
  ViewBmp: TBitmap;
  Elements: IHTMLElementCollection;
  Element: IHTMLElement;
  ParentElement: IHTMLElement;
  tmpElement: IDispatch;
  ViewObject: IViewObject;
  i: Integer;
begin
  if (not IstBody) then
    Exit;

  KeepAlign := Align;
  Align := alNone;
  try
    X := GetSystemMetrics(SM_CXVSCROLL);
    Y := GetSystemMetrics(SM_CYHSCROLL);

    P := 0;
    Q := 0;

    if (wo3D in FOptions) then begin
      P := GetSystemMetrics(SM_CXEDGE);
      Q := GetSystemMetrics(SM_CYEDGE);
    end;

    vBmp.Width := OleObject.document.body.scrollwidth + X + P * 2;
    vBmp.Height := OleObject.document.body.scrollHeight + Y + Q * 2;

    if (vWidth > 0) then
      vBmp.Width := vWidth;

    if (vHeight > 0) then
      vBmp.Height := vHeight;

    OrgWidth := Width;
    OrgHeight := Height;

    OrgPos.X := OleObject.document.body.scrollLeft;
    OrgPos.Y := OleObject.document.body.scrollTop;

    OleObject.document.body.scrollLeft := 0;
    OleObject.document.body.scrollTop := 0;

    SetBounds(Left, Top, vBmp.Width, vBmp.Height);

    // Body
    OleDraw(
      ControlInterface,
      DVASPECT_DOCPRINT,
      vBmp.Canvas.Handle,
      Rect(0, 0, vBmp.Width, vBmp.Height));

    // Flash
    if
      (Document <> nil) and
      (Document.QueryInterface(IHTMLDocument2, HTMLDoc) = S_OK)
    then begin
      Elements := HTMLDoc.all;

      ViewBmp := TBitmap.Create;
      try
        ViewBmp.Width := vBmp.Width;
        ViewBmp.Height := vBmp.Height;

        for i := 0 to Elements.length - 1 do begin
          tmpElement := Elements.item(i, i);

          if (tmpElement.QueryInterface(IHTMLElement, Element) <> S_OK) then
            Continue;

          if
            (tmpElement.QueryInterface(IViewObject, ViewObject) = S_OK)
          then begin
            S := 0;
            T := 0;

            ParentElement := Element;

            while (ParentElement <> nil) do begin
              Inc(S, ParentElement.offsetLeft);
              Inc(T, ParentElement.offsetTop);

              ParentElement := ParentElement.offsetParent;
            end;

            OleDraw(
              ViewObject,
              DVASPECT_CONTENT,
              ViewBmp.Canvas.Handle,
              Rect(0, 0, Element.offsetWidth, Element.offsetHeight));

            BitBlt(
              vBmp.Canvas.Handle,
              S,
              T,
              Element.offsetWidth,
              Element.offsetHeight,
              ViewBmp.Canvas.Handle,
              0,
              0,
              SRCCOPY);
          end;
        end;
      finally
        ViewBmp.Free;
      end;
    end;

    vBmp.Canvas.Draw(-P, -Q, vBmp);

    vBmp.Width := vBmp.Width - X - P * 2;
    vBmp.Height := vBmp.Height - Y - Q * 2;

    SetBounds(Left, Top, OrgWidth, OrgHeight);

    Forms.Application.ProcessMessages;

    while (IUnknown(OleObject.document.body) = nil) do
      Forms.Application.ProcessMessages;

    while (OleObject.document.body.scrollLeft <> OrgPos.X) do begin
      OleObject.document.body.scrollLeft := OrgPos.X;
      Forms.Application.ProcessMessages;
    end;

    while (OleObject.document.body.scrollTop <> OrgPos.Y) do begin
      OleObject.document.body.scrollTop := OrgPos.Y;
      Forms.Application.ProcessMessages;
    end;
  finally
    Align := KeepAlign;
  end;
end;

initialization
begin
  CoInitialize(nil);
end;

finalization
begin
  CoUninitialize;
end;

end.
